/*
* Creates a first partition consisting on groups of worlds with same labeling
*/

firstPartition(Worlds,WorldsFirstPartition):-
    findall(Label,member(w(_,Label),Worlds),Labels),
    list_to_set(Labels,LabelSet),
    group_by_label(LabelSet,Worlds,WorldsFirstPartition).
    
/*
* Creates groups of worlds with same label.
*/
group_by_label([],_,[]):-!.
group_by_label([Label|MoreLabels],Worlds,[GroupOfLabel|MoreGroups]):-
    findall(w(Id,Label),member(w(Id,Label),Worlds),GroupOfLabel),
    group_by_label(MoreLabels,Worlds,MoreGroups).

/*
* refine_partition(Blocks,OldPartition,Agents,Edges,NewPartition)
* A partition is a list of lists where each list contains Kripke model states
* Edges define the relation between states on agents
*/    
/* Stop when no change in partitions*/
refine_partition([],_,_,_,[]):-!.

refine_partition([Block|MoreBlocks],OldPartition,Agents,Edges,NewPartition):-
    block_partition(Block,OldPartition,Agents,Edges,BlockPartition),
    refine_partition(MoreBlocks,OldPartition,Agents,Edges,NewMoreBlocksPartition),
    append(BlockPartition,NewMoreBlocksPartition,NewPartition).
    

/*
* block_partition(Block,PrevPartition,BlockPartition)
* This predicate partitions a block
*/

block_partition([World|MoreWorlds],PrevPartition,Agents,Edges,BlockPartition):-
    compatible_states(World,MoreWorlds,PrevPartition,Agents,Edges,CompatibleStates),
    subtract(MoreWorlds,CompatibleStates,IncompatibleStates),
    (
        (
            IncompatibleStates = [],
            BlockPartition = [[World|MoreWorlds]],
            !
        );
        (
            block_partition(IncompatibleStates,PrevPartition,Agents,Edges,IncompatibleStatesPartition),
            append([[World|CompatibleStates]],IncompatibleStatesPartition,BlockPartition)
        )
    ).

/*
* Computes the compatible states 
* go to same blocks with every agent
*/
compatible_states(World,ToTest,PrevPartition,Agents,Edges,CompatibleStates):-
    findall(State,(member(State,ToTest),compatible(World,State,PrevPartition,Agents,Edges)),CompatibleStates ).
    
/*
* World1 and World2 go to same blocks relative to Partition on the relation given by Edges for each agent 
*/

/* All worlds are compatible relative to no agent */
compatible(_,_,_,[],_):-!.

compatible(w(World1,_),w(World2,_),Partition,[Agent|MoreAgents],Edges):-
    filterEdgesByAnyAgent(Edges,[Agent],Filtered),
    findall(Tgt1,(member(edge(World1,_,Tgt1),Filtered)),WorldsFrom1),
    blocks(WorldsFrom1,Partition,BlocksFrom1),
    findall(Tgt2,(member(edge(World2,_,Tgt2),Filtered)),WorldsFrom2),
    blocks(WorldsFrom2,Partition,BlocksFrom2),
    set_equal(BlocksFrom1,BlocksFrom2),
    compatible(w(World1,_),w(World2,_),Partition,MoreAgents,Edges).
    
/*
* Retrives the Blocks from Partition wich contains any world on Worlds
*/

blocks(Worlds,Partition,Blocks):-
    findall(Block,(member(Block,Partition),member(World,Worlds),member(w(World,_),Block)),Blocks).
    
/*
* Refines Partition until there is no change.
*/
refine_partition_until_equal(Partition,Agents,Edges,FinalPartition):-
    refine_partition(Partition,Partition,Agents,Edges,NewPartition),
    (
        (
            Partition = NewPartition,
            FinalPartition = Partition,
            !
        );
        (
            refine_partition_until_equal(NewPartition,Agents,Edges,FinalPartition)
        )
    ).

/*
* minimize(model(Worlds,Edges),model(NewWolrds,NewEdges))
* Minimize the model by bisimulation.
*/
minimize(model(Worlds,Edges),model(NewWorlds,NewEdges)):-
    firstPartition(Worlds,InitialPartition),
    agents_from_edges(Edges,Agents),
    refine_partition_until_equal(InitialPartition,Agents,Edges,FinalPartition),
    partition_to_worlds(FinalPartition,NewWorlds),
    rebuild_edges(Edges,NewWorlds,NewEdges).
    
/*
* Converts a partition to a set of worlds 
* NOTICE all worlds on a block must have the same labeling
*/

partition_to_worlds([],[]):-!.

partition_to_worlds([Block|MorBlocks],[World|MoreWorlds]):-
    block_to_world(Block,World),
    partition_to_worlds(MorBlocks,MoreWorlds).
    
/*
* block_to_world(Block,World)
* being block a set of worlds
* Converts a block to a world
*/

block_to_world([w(Id,Label)|MoreWorlds],w(AllWorldIds,Label)):-
    findall(WorldId,member(w(WorldId,_),[w(Id,Label)|MoreWorlds]),AllWorldIds).

/*
* rebuild_edges(Edges,Partition,NewEdges)
* Rebuilds Edges according to partition creating NewEdges
*/
rebuild_edges(Edges,NewWorlds,NewEdges):-
    substitute_worlds_for_blocks(Edges,NewWorlds,BlockEdges),
    fuse_edges(BlockEdges,[],NewEdges).

/*
* substitute_worlds_for_blocks(Edges,NewWorlds,NewEdges)
* Substitute the worlds on edges by the block containing the world.
* New worlds have the form w( [Id1,Id2,..,Idn] , Label) 
* [Id1,Id2,..,Idn] corresponding to a list of old world ids
*/

substitute_worlds_for_blocks([],_,[]):-!.
    
substitute_worlds_for_blocks([edge(World1,Label,World2)|MoreEdges],NewWorlds,[edge(Block1,Label,Block2)|MoreNewEdges]):-
    member(w(Block1,_),NewWorlds),
    member(World1,Block1),
    member(w(Block2,_),NewWorlds),
    member(World2,Block2),
    substitute_worlds_for_blocks(MoreEdges,NewWorlds,MoreNewEdges).

/*
* Fuse the edges with same source and target with an union on the Labels
*/    
fuse_edges([],NewEdges,NewEdges):-!.

fuse_edges([edge(Src,Label,Tgt)|MoreEdges],FusedEdges,NewEdges):-
    (
        (
            member(edge(Src,OldLabel,Tgt),FusedEdges),
            union(Label,OldLabel,FusedLabel),
            replace(FusedEdges,edge(Src,OldLabel,Tgt),edge(Src,FusedLabel,Tgt),NewFusedEdges),
            fuse_edges(MoreEdges,NewFusedEdges,NewEdges),
            !
        );
        (
            fuse_edges(MoreEdges,[edge(Src,Label,Tgt)|FusedEdges],NewEdges)
        )
    ).



