/* 
* Calculates the graphviz code to represent a model of epistemic logic.
* Columns is the number of columns the memory will be placed.
*/

makeModel(model(Worlds,Arrows),Columns,Str):-
    makeWorlds(Worlds,Columns,WStrAL),
    makeArrows(Arrows,EStrAL),
    append([[s,t,r,i,c,t,' ',/*d,i,*/g,r,a,p,h,'{','\n'],WStrAL,EStrAL,['}']],StrAL),
    string_chars(Str,StrAL).

makeSimplifiedModel(model(Worlds,Arrows),Columns,Str):-
    makeWorlds(Worlds,Columns,WStrAL),
    remove_reflexivity(Arrows,ArrowsWithoutRef),
    remove_transitivity(ArrowsWithoutRef,ArrowsWithoutRefNorTrans),
    remove_direction(ArrowsWithoutRefNorTrans,ArrowsWithoutRefNorDirNorTrans),
    makeArrowsAsEdges(ArrowsWithoutRefNorDirNorTrans,EStrAL),
    append([[g,r,a,p,h,'{','\n'],WStrAL,EStrAL,['}']],StrAL),
    string_chars(Str,StrAL).

/*
* Uses makeModel to represent an extended action model, representing the new information as the labeling function in Kripke models.
* The precondition function is omitted for text description outside the graph.
* Calculates the graphviz code to represent a model of epistemic logic.
* Columns is the number of columns the memory will be placed.
*/
makeExtendedActionModel(extended_action_model(ExtendedActionPoints,Arrows),Columns,Str):-
    extendedActionPointsToWorlds(ExtendedActionPoints,Worlds),
    makeModel(model(Worlds,Arrows),Columns,Str).

makeSimplifiedExtendedActionModel(extended_action_model(ExtendedActionPoints,Arrows),Columns,Str):-
    extendedActionPointsToWorlds(ExtendedActionPoints,Worlds),
    makeSimplifiedModel(model(Worlds,Arrows),Columns,Str).

/*
* Converts an extended action point to a wold (in structure) omitting the pre condition function and 
* representing the labelling function in the extended action model as the labelling function in a Kripke model.
*/

extendedActionPointsToWorlds([],[]):-!.

extendedActionPointsToWorlds([eap(Id,_,NewInfo)|MoreEAPs],[w(Id,NewInfo)|MoreWorlds]):-
    extendedActionPointsToWorlds(MoreEAPs,MoreWorlds).


/*
* Represents the edges as dot language
*/    

makeArrows([],[]):-!.

makeArrows([Arrow|MoreArrows],Res):-
    makeArrow(Arrow,ArrowAL),
    makeArrows(MoreArrows,MoreArrowsAL),
    append(ArrowAL,MoreArrowsAL,Res).

makeArrow(edge(Id1,Labels,Id2),StrAL):-
    term_string(Id1,Id1Str_),
    string:concat(Id1Str_,"\"",Id1Str),
    term_string(Id2,Id2Str_),
    string:concat(Id2Str_,"\"",Id2Str),
    string_concat("\"",Id1Str,Tmp1),
    string_concat(Tmp1,"--\"",Tmp2),
    string_concat(Tmp2,Id2Str,Tmp3),
    string_concat(Tmp3," [fontsize = 30 , label=\"",Tmp4),
    list_to_string(Labels,0,1,LabelsStr),
    string_concat(Tmp4,LabelsStr,Tmp5),
    string_concat(Tmp5,"\"];\n",All),
    string_chars(All,StrAL).

makeArrowsAsEdges([],[]):-!.

makeArrowsAsEdges([Arrow|MoreArrows],Res):-
    makeArrowAsEdge(Arrow,ArrowAL),
    makeArrowsAsEdges(MoreArrows,MoreArrowsAL),
    append(ArrowAL,MoreArrowsAL,Res).

makeArrowAsEdge(edge(Id1,Labels,Id2),StrAL):-
    term_string(Id1,Id1Str_),
    string:concat(Id1Str_,"\"",Id1Str),
    term_string(Id2,Id2Str_),
    string:concat(Id2Str_,"\"",Id2Str),
    string_concat("\"",Id1Str,Tmp1),
    string_concat(Tmp1,"--\"",Tmp2),
    string_concat(Tmp2,Id2Str,Tmp3),
    string_concat(Tmp3," [fontsize = 30 , label=\"",Tmp4),
    list_to_string(Labels,0,1,LabelsStr),
    string_concat(Tmp4,LabelsStr,Tmp5),
    string_concat(Tmp5,"\"];\n",All),
    string_chars(All,StrAL).
    

makeWorlds([],_,[]):-!.

makeWorlds([World|MoreWorlds],Columns,Res):-
    makeWorld(World,Columns,WorldAL),
    makeWorlds(MoreWorlds,Columns,MoreWorldsAL),
    append(WorldAL,MoreWorldsAL,Res).

makeWorld(w(Id,Values),Columns,WorldAL):-
    term_string(Id,IdStr),
    %getVars(0,Values,Vars),
    list_to_string(Values,Columns,1,VarsStr),
    string_chars(VarsStr,VarChars),
    delete(VarChars,'"',VarChars2),
    string_chars(VarsStr2,VarChars2),
    string_concat("(",IdStr,Tmp0),
    string_concat(Tmp0,")\\n",Tmp),
    string_concat(Tmp,VarsStr2,Label),
    string_concat("\"",IdStr,IdStr_),
    string_concat(IdStr_,"\"",IdStr__),
    string_concat(IdStr__," [fontsize = 30, label=\"",Tmp1),
    string_concat(Tmp1,Label,Tmp2),
    string_concat(Tmp2,"\"];\n",All),
    string_chars(All,WorldAL).

list_to_string([],_,_,""):-!.

list_to_string([X],_,_,XS):-
    term_string(X,XS),
    !.

list_to_string([X|More],Modulo,Column,Str):-
    term_string(X,XS),
    string_concat(XS,",",Tmp_),
    (
        (
            Modulo = Column,
            string_concat(Tmp_,"\n",Tmp),
            NewColumn = 1,
            !
        );
        (   Tmp = Tmp_,
            NewColumn is Column + 1
        )
    ),
    list_to_string(More,Modulo,NewColumn,MoreStr),
    string_concat(Tmp,MoreStr,Str),
    !.
    
getVars(_,[],[]):-!.

getVars(Index,[1|MoreValues],[Var|MoreVars]):-
    term_string(Index,IndexStr),
    string_concat("p",IndexStr,Var),
    Index2 is Index + 1,
    getVars(Index2,MoreValues,MoreVars),
    !.

getVars(Index,[0|MoreValues],MoreVars):-
    Index2 is Index + 1,
    getVars(Index2,MoreValues,MoreVars).

makeComplex(Complex,NColumns,Colors,Str):-
    findall(Vertex,(member(Vertex,Complex),length(Vertex,1)),Vertices),
    verticesWithId(Vertices,0,VerticesWithId),
    makeVertices(VerticesWithId,NColumns,Colors,VerticesStr),
    findall(Edge,(member(Edge,Complex),length(Edge,2)),Edges),
    makeEdges(Edges,VerticesWithId,EdgesStr),
    string_concat("graph{\n",VerticesStr,Tmp1),
    string_concat(Tmp1,EdgesStr,Tmp2),
    string_concat(Tmp2,"}\n",Str).

/*
* Need ids to build the graph.
*/

verticesWithId([],_,[]).

verticesWithId([Vertex|MoreVertices],Id,[v(Id,Vertex)|MoreVerticesWithId]):-
    Id2 is Id + 1,
    verticesWithId(MoreVertices,Id2,MoreVerticesWithId).

makeVertices([],_,_,"").

makeVertices([v(Id,[v(ColorOnComplex,View)])|MoreVertices],Columns,Colors,Str):-
    member( ColorOnComplex => Color , Colors),
    term_string(Id,VStr__),
    string_concat("\"",VStr__,VStr_),
    string_concat(VStr_,"\"",VStr),
    string_concat(VStr, " [fontsize = 30, style = filled,fillcolor = ",Tmp1),
    string_concat(Tmp1, Color,Tmp2),
    list_to_string(View,Columns,1,ViewS__),
    string_concat("[",ViewS__,ViewS_),
    string_concat(ViewS_,"]",ViewS),
    %term_string(View,ViewS),
    string_chars(ViewS,ViewChars),
    scapeSpecial(ViewChars,ViewScapped),
    string_chars(ViewStr,ViewScapped),
    string_concat(Tmp2,", label = \"",Tmp3),
    string_concat(Tmp3,ViewStr,Tmp4),
    string_concat(Tmp4,"\"",Tmp5),
    string_concat(Tmp5, "];\n",VertexStr),
    makeVertices(MoreVertices,Columns,Colors,MoreVerticesStr),
    string_concat(VertexStr,MoreVerticesStr,Str).

scapeSpecial([],[]):-!.

scapeSpecial([Char|MoreChars],['\\'|[Char|MoreCharsScaped]]):-
    member(Char,['"','\\']),
    scapeSpecial(MoreChars,MoreCharsScaped).

scapeSpecial([Char|MoreChars],[Char|MoreCharsScaped]):-
    scapeSpecial(MoreChars,MoreCharsScaped).
    

makeEdges([],_,"").

makeEdges([[Vertex1,Vertex2]|MoreEdges],VerticesWithId,Str):-
    member(v(Id1,[Vertex1]),VerticesWithId),
    term_string(Id1,Vertex1Str__),
    string_concat("\"",Vertex1Str__,Vertex1Str_),
    string_concat(Vertex1Str_,"\"",Vertex1Str),
    member(v(Id2,[Vertex2]),VerticesWithId),
    term_string(Id2,Vertex2Str__),
    string_concat("\"",Vertex2Str__,Vertex2Str_),
    string_concat(Vertex2Str_,"\"",Vertex2Str),
    string_concat(Vertex1Str," -- ",Tmp1),
    string_concat(Tmp1,Vertex2Str,Tmp2),
    string_concat(Tmp2,";\n",Tmp3),
    makeEdges(MoreEdges,VerticesWithId,MoreEdgesStr),
    string_concat(Tmp3,MoreEdgesStr,Str).

/*
* remove_reflexivity(Edges,NewEdges)
* Calculates the set NewEdges subset of Edges so that
* there is no edge from and to the same vertex.
*/

remove_reflexivity(Edges,NewEdges):-
    findall(edge(V,_,V),member(edge(V,_,V),Edges),ReflexiveEdges),
    subtract(Edges,ReflexiveEdges,NewEdges).

/*
* remove_direction(Edges,NewEdges)
* Calculates the set NewEdges subset of Edges so that
* if edge(V1,_V2) in NewEdges then edge(V2,_V1) not in NewEdges
*/

remove_direction([],[]):-!.

remove_direction([edge(Origin,Label,Target)|MoreEdges],[edge(Origin,Label,Target)|NewMoreEdgesWithoutCurrent]):-
    delete(MoreEdges,edge(Target,Label,Origin),MoreEdgesWithoutCurrent),
    remove_direction(MoreEdgesWithoutCurrent,NewMoreEdgesWithoutCurrent).

/*
* remove_transitivity(Edges,NewEdges)
* Calculates NewEdges subset of Edges so that
* there is no transitivity in NewEdges.
*/

remove_transitivity(Edges,NewEdges):-
    member(edge(Id1,Label1,Id2),Edges),
    member(edge(Id2,Label2,Id3),Edges),
    dif(Id1,Id3),
    intersection(Label1,Label2,LabelInt),
    dif(LabelInt,[]),
    member(edge(Id1,Label3,Id3),Edges),
    intersection(LabelInt,Label3,LabelIntersection),
    dif(LabelIntersection,[]),
    (
        (
            contains(LabelIntersection,Label3),
            delete(Edges,edge(Id1,Label3,Id3),Edges2),
            delete(Edges2,edge(Id3,Label3,Id1),Edges3),
            remove_transitivity(Edges3,NewEdges),
            !
        );
        (
            %trace,
            subtract(Label3,LabelIntersection,NewLabel),
            delete(Edges,edge(Id1,Label3,Id3),Edges2),
            delete(Edges2,edge(Id3,Label3,Id1),Edges3),
            remove_transitivity([edge(Id1,NewLabel,Id3)|[edge(Id3,NewLabel,Id1)|Edges3]],NewEdges)
        )
    ),
    !.

/*
* Once there is no other transitive arrow I have finished.
*/
remove_transitivity(Edges,Edges).

/* Predicates to call graphviz */

/*
* generate_model(Model,Columns,Program,DotFile,GifFile)
* generates .dot file and gif file using Program
*/

generate_model(Model,Columns,Program,DotFile,GifFile):-
    makeModel(Model,Columns,Str),
    current_output(Orig), % save current output
    open(DotFile, write, Out),
    set_output(Out),
    write(Str),
    close(Out),
    set_output(Orig),
    generate_graph(Program,DotFile,GifFile),
    !.

generate_simplified_model(Model,Columns,Program,DotFile,GifFile):-
    makeSimplifiedModel(Model,Columns,Str),
    current_output(Orig), % save current output
    open(DotFile, write, Out),
    set_output(Out),
    write(Str),
    close(Out),
    set_output(Orig),
    generate_graph(Program,DotFile,GifFile),
    !.

generate_complex(Complex,NColumns,Colors,Program,DotFile,GifFile):-
    makeComplex(Complex,NColumns,Colors,Str),
    current_output(Orig), % save current output
    open(DotFile, write, Out),
    set_output(Out),
    write(Str),
    close(Out),
    set_output(Orig),
    generate_graph(Program,DotFile,GifFile),
    !.

generate_graph(Program,DotFile,GifFile):-
    term_string(Program,ProgramStr),
    string_concat(ProgramStr," -Tpng ",Tmp0),
    term_string(DotFile,DotFileStr),
    string_concat(Tmp0,DotFileStr,Tmp1),
    string_concat(Tmp1," -o ",Tmp2),
    term_string(GifFile,GifFileStr),
    string_concat(Tmp2,GifFileStr,CmdStr),
    atom_string(Atom,CmdStr),
    shell(Atom).

generate_extended_action_model(ActionModel,Columns,Program,DotFile,File):-
    makeExtendedActionModel(ActionModel,Columns,Str),
    current_output(Orig), % save current output
    open(DotFile, write, Out),
    set_output(Out),
    write(Str),
    close(Out),
    set_output(Orig),
    generate_graph(Program,DotFile,File),
    !.

generate_simplified_extended_action_model(ActionModel,Columns,Program,DotFile,File):-
    makeSimplifiedExtendedActionModel(ActionModel,Columns,Str),
    current_output(Orig), % save current output
    open(DotFile, write, Out),
    set_output(Out),
    write(Str),
    close(Out),
    set_output(Orig),
    generate_graph(Program,DotFile,File),
    !.
    
    
