/*
* reaches(W1,Reached,Edges)
* true if Edges contains a path from W1 to all worlds on Reached
*/

reaches(WId,Reached,Edges):-
    bfs([WId],[],Edges,Reached).

/*
* Runs a bfs search
* bfs(Wid,Open,Closed,Edges,Visited)
*/

/*
* Once all nodes have been closed, we know all reached are the closed set.
*/
bfs([],Closed,_,Closed):-!.

/*
* Tail recursion for bfs
*/

bfs([World|MoreWorlds],Closed,Edges,AllVisited):-
    neighbors(World,Edges,Neighbors),
    append([World|MoreWorlds],Closed,Visited),
    subtract(Neighbors,Visited,Reached),
    append(MoreWorlds,Reached,Opened),
    bfs(Opened,[World|Closed],Edges,AllVisited).

/*
* Calculates the neighbors of a world given the edges 
*/

neighbors(WId,Edges,Neighbors):-
    findall(WId2,member(edge(WId,_,WId2),Edges),Neighbors).

filterEdgesByAnyAgent(Edges,Agents,Filtered):-
    findall(
        edge(WId,ELabel,WId2),
        (
            member(edge(WId,ELabel,WId2),Edges),
            not(intersection(ELabel,Agents,[]))
        ),
        Filtered).

filterEdgesByAllAgents(Edges,Agents,Filtered):-
    findall(
        edge(WId,ELabel,WId2),
        (
            member(edge(WId,ELabel,WId2),Edges),
            contains(ELabel,Agents)
        ),
        Filtered).


/*
* The atomic formulas here are just p_Index_Value
* p(Index,Value) where Index > 0 
* p(Index,Value) is true iff the memory on Index position has value Value
* As sintax sugar being Round and Agent the number of round and agent respectively.
* p(Round,Agent,Value) is true iff the memory on index Roud * Agent has value Velue.
*/

/*
* Defining operators
*/

:-op(800,fx,'not').

:-op(900,xfy,'and').

:-op(900,xfy,'or').

:-op(1000,xfy,'=>').

:-op(1000,yfx,'<=>').

/*
* Calculates the maximum length of the list of variables for all Worlds.
* The model must have at least a world.
*/
at_vars([w(_,Values)],CL):-
    length(Values,CL).

at_vars([w(_,Values)|MoreWorlds],N):-
    length(Values,CL),
    at_vars(MoreWorlds,ML),
    N is max(CL,ML).

/*
* Gets the agents from the edges.
* Asumes the labelling is a set 
*/
agents([],[]).

agents([edge(_,Label,_)|MoreEdges],Agents):-
    agents(MoreEdges,MoreAgents),
    union(Label,MoreAgents,Agents).

/*
* checks a formula on the given model.
*/
check(Formula,World,Model):-
    (
        (
            formula(Formula,Model),
            !
        );
        representation_error('Not a formula...')
    ),
    valid(Formula,World,Model),!.


/* Here we define what a formula is
* 0 is bottom (false)
* 1 is top (true)
*/


formula(0,_):-!.

formula(1,_):-!.

formula(p(Index),_):-
    integer(Index),
    Index >= 0,
    !.


formula(p(Index,_),_):-
    integer(Index),
    Index >= 0,
    !.
    
formula(p(NAgents,Round,Agent,_),_):-
    integer(NAgents),    
    integer(Round),
    integer(Agent),
    NAgents >= 0,
    Round >= 0,
    Agent >= 0,
    !.

formula(not F,Model):-
    formula(F,Model).

formula(F1 and F2,Model):-
    formula(F1,Model),
    formula(F2,Model).

formula(F1 or F2,Model):-
    formula(F1,Model),
    formula(F2,Model).

formula(F1 => F2,Model):-
    formula(F1,Model),
    formula(F2,Model).

formula(F1 <=> F2,Model):-
    formula(F1,Model),
    formula(F2,Model).

formula(k(Agent,Formula),model(Worlds,Edges)):-
    agents(Edges,Agents),
    member(Agent,Agents),
    formula(Formula,model(Worlds,Edges)).

formula(e(Agents,Formula),model(Worlds,Edges)):-
    agents(Edges,AllAgents),
    contains(AllAgents,Agents),
    formula(Formula,model(Worlds,Edges)).

formula(c(Agents,Formula),model(Worlds,Edges)):-
    agents(Edges,AllAgents),
    contains(AllAgents,Agents),
    formula(Formula,model(Worlds,Edges)).

formula(d(Agents,Formula),model(Worlds,Edges)):-
    agents(Edges,AllAgents),
    contains(AllAgents,Agents),
    formula(Formula,model(Worlds,Edges)).

/*
* for K E C and D we need functors 
* k(Agent,Formula)
* e(G,Formula)
* c(G,Formula)
* d(G,Formula)
*/

valid(1,World,model(Worlds,_)):-
    member(w(World,_),Worlds).

valid(p(Index),World,model(Worlds,_)):-
    member(w(World,Values),Worlds),
    ith(Values,Index,Value),
    Value = 1.
    
valid(p(Index,Value),World,model(Worlds,_)):-
    member(w(World,Values),Worlds),
    ith(Values,Index,Value).
    
valid(p(NAgents,Round,Agent,Value),World,Model):-
    Index is (NAgents * Round) + Agent,
    valid(p(Index,Value),World,Model).

valid( not Formula , World , Model ) :-
    not(valid(Formula,World,Model)).

valid(Formula1 and Formula2, World, Model ):-
    valid(Formula1,World,Model),
    valid(Formula2,World,Model).

valid(Formula1 or Formula2, World, Model ):-
    (valid(Formula1,World,Model),!);
    valid(Formula2,World,Model).

valid(Formula1 => Formula2, World, Model ):-
    valid(not Formula1 or Formula2,World,Model).

valid( Formula1 <=> Formula2, World, Model ):-
    valid(Formula1 => Formula2,World,Model),
    valid(Formula2 => Formula1,World,Model).

valid(k(Agent,Formula),World,model(Worlds,Edges)):-
    filterEdgesByAnyAgent(Edges,[Agent],Filtered),
    neighbors(World,Filtered,Neighbors),
    checkFormulaOnWorlds(Neighbors,Formula,model(Worlds,Edges)).

valid(e(Agents,Formula),World,Model):-
    allKnow(Agents,Formula,World,Model).

valid(c(Agents,Formula),World,model(Worlds,Edges)):-
    filterEdgesByAnyAgent(Edges,Agents,Filtered),
    reaches(World,Reached,Filtered),
    checkFormulaOnWorlds(Reached,Formula,model(Worlds,Edges)).

valid(d(Agents,Formula),World,model(Worlds,Edges)):-
    filterEdgesByAllAgents(Edges,Agents,Filtered),
    neighbors(World,Filtered,Neighbors),
    checkFormulaOnWorlds(Neighbors,Formula,model(Worlds,Edges)).
    

/*
* allKnow(Agents,Formula,World,Model)
* ture if all agents knows Formula in World of Model.
*/

allKnow([],_,_,_):-!.

allKnow([Agent|MoreAgents],Formula,World,Model):-
    valid(k(Agent,Formula),World,Model),
    allKnow(MoreAgents,Formula,World,Model).
    
    
/*
* checkFormulaOnWorlds(Worlds,Formula,Model)
* True if Formula is valid in all worlds of Model
*/

checkFormulaOnWorlds([],_,_):-!.

checkFormulaOnWorlds([World|MoreWorlds],Formula,Model):-
    valid(Formula,World,Model),
    checkFormulaOnWorlds(MoreWorlds,Formula,Model).



