/*
* This predicate calculates bottom as '_' when two values differ.
* When they're equal the mixed value is the same.
* mix(A,B,Mixed)
*/

mix(A,B,A):- A = B,!.
mix(_,_,'_').

/*
* mix_list(L1,L2,ML)
* Uses mix to create a concrete view
*/
mix_list([],[],[]):-!.
mix_list([V1|More1],[V2|More2],[Mixed|MoreMixed]):-
    mix(V1,V2,Mixed),
    mix_list(More1,More2,MoreMixed).

/*
*create_concrete_view(ListOfListOfValues,ConcreteView)
* Creates a concrete view from a list of lists of values
*/
create_concrete_view([Values],Values):-!.

create_concrete_view([Values|MoreValues],View):-
    create_concrete_view(MoreValues,PreView),
    mix_list(PreView,Values,View).


/*
* model_to_complex(Model,Complex)
* Given a kripke model, calculates a simplicial model associated.
*/
model_to_complex(Model,Complex):-
    model_to_simplices_and_complex(Model,_,Complex).

model_to_simplices(Model,Simplices):-
    model_to_simplices_and_complex(Model,Simplices,_).

model_to_simplices_and_complex(model(Worlds,Edges),Simplices,Complex):-
    calculate_complex(Worlds,Edges,Worlds,Simplices,Complex).


/*
* edges_of_world(World,AllEdges,EdgesIncluding)
* Given a particular world, it takes the subset of Edges where it is origin.
*/

edges_of_world(_,[],[]):-!.

edges_of_world(World,[edge(Origin,Label,Target)|MoreEdges],EdgesIncludingWorld):-
    edges_of_world(World,MoreEdges,MoreEdgesIncludingWorld),
    (
        (
            World = Origin,
            EdgesIncludingWorld = [edge(Origin,Label,Target)|MoreEdgesIncludingWorld],
            !
        );
        (
            EdgesIncludingWorld = MoreEdgesIncludingWorld
        )
    ).

/*
* calculate_complex(Worlds,Edges,AllWorlds,Simplices,Complex)
* Calculates the complex corresponding to a model with the given Worlds and Edges.
*/

calculate_complex([],_,[],[]):-!.

calculate_complex([w(World,_)|MoreWorlds],AllEdges,Worlds,Simplices,Complex):-
    calculate_complex(MoreWorlds,AllEdges,Worlds,PreSimplices,PreComplex),
    calculate_simplex(World,AllEdges,Worlds,Simplex,PSimplex),
    union(Simplex,PreSimplices,Simplices),
    union(PSimplex,PreComplex,Complex).

/*
* calculate_simplex(World,Edges,Worlds,Simplex,PSimplex)
* Given a World in Worlds and the Edges in the Kripke model,
* calculates the corresponding Simplex  and all the subsimplices in the corresponding simplicial complex.
*/

calculate_simplex(World,AllEdges,Worlds,Simplex,PSimplex):-
    edges_of_world(World,AllEdges,Edges),
    get_colors_from_edges(Edges,Colors),
    calculate_vertices(Colors,Edges,Worlds,Simplex),
    oset_power(Simplex,PSimplex).

/*
* calculate_vertices(Colors,Edges,Vertices)
* Calculates vertices with colors in Colors and View calculed from Edges
*/
calculate_vertices([],_,_,[]):-!.

calculate_vertices([Color|MoreColors],Edges,Worlds,[Vertex|MoreVertices]):-
    calculate_vertex(Color,Edges,Worlds,Vertex),
    calculate_vertices(MoreColors,Edges,Worlds,MoreVertices).


/* Calculates a vertex given a color and the edges where the values will be gotten*/
calculate_vertex(Color,Edges,Worlds,v(Color,View)):-
    get_label_values(Color,Edges,Worlds,Values),
    create_concrete_view(Values,View).
    

/*Given the set of edges from a Kripke model calculates the colors to create the simplicial modell*/
get_colors_from_edges(Edges,Colors):-
    labellings(Edges,Labellings),
    max_length(Labellings,Max),
    findall(Labelling,(member(Labelling,Labellings),length(Labelling,Max)),[Colors|_]).

/*
* Given the set of edges from a Kripke model calculates the list of Labellings
*/
labellings([],[]):-!.

labellings([edge(_,L,_)|MoreEdges],[L|MoreLabels]):-
    labellings(MoreEdges,MoreLabels).

/* 
* get_label_values(Label,Edges,Worlds,Values)
* Gets target world's values when the edge is labelled by L
*/

get_label_values(_,[],_,[]):-!.

get_label_values(L,[edge(_,Ls,Target)|MoreEdges],Worlds,[Values|MoreEdgesFiltered]):-
    member(L,Ls),
    member(w(Target,Values),Worlds),
    get_label_values(L,MoreEdges,Worlds,MoreEdgesFiltered),!.

get_label_values(L,[_|MoreEdges],Worlds,MoreEdgesFiltered):-
    get_label_values(L,MoreEdges,Worlds,MoreEdgesFiltered),!.
